/*
 *	NEWS-0.4.6 - GTK-based RSS Ticker - Copyright (C) Emmanuel Thomas-Maurin 2009-2011
 *	<manutm007@gmail.com>
 *
 * 	This program is free software: you can redistribute it and/or modify
 * 	it under the terms of the GNU General Public License as published by
 * 	the Free Software Foundation, either version 3 of the License, or
 * 	(at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * 	GNU General Public License for more details.
 *
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define APP_VERSION_NUMBER	"0.4.6"
#define APP_WEBSITE		"http://www.newsrssticker.com"
#define APP_SUPPORT_WEBSITE	"http://www.newsrssticker.com/help.php"
#define APP_DOWNLOAD_WEBSITE	"http://www.newsrssticker.com/download.php"
#define APP_SUPPORT_EMAIL	"newsrssticker@gmail.com"/*"contact@newsrssticker.com"*/

#define _GNU_SOURCE
#define _ISOC99_SOURCE

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#ifndef G_OS_WIN32
# include <grp.h>
# include <pwd.h>
#else
# include <windows.h>
# include <rpcdce.h>
# include <iphlpapi.h>
# include <winreg.h>
# include <shlobj.h>
#endif
#ifndef G_OS_WIN32
# include <glib-2.0/glib/gstdio.h>
#else
# include <../glib-2.0/glib/gstdio.h>
#endif
#include <libxml/parser.h>
#include <libxml/tree.h>
#include "../libetm-0.3/libetm.h"

#define	USE_GUI			TRUE	/* to implement or not gui/menu features at compile time
					 * otherwise, cli only app */
#if USE_GUI
# define RESOURCE_NONE_QUIT	FALSE	/* to quit or not if no resource id specified */
# define OPTION_ERROR_QUIT	FALSE	/* to quit or not if unknown or invalid option(s) */
#else
# define RESOURCE_NONE_QUIT	TRUE
# define OPTION_ERROR_QUIT	TRUE
#endif

#define MENU_V			TRUE	/* setting widget layout -> vertical or horizontal menu
					 * [true = vertical] seems a good choice */
#define MENU_H			!MENU_V	/* otherwise, deprecated buggy stuff (no support for undecorated win, etc) */
#define MENU_LAB		"?"	/* root h menu label */

#define DRWA_WIDTH_MIN		100
#define DRWA_HEIGHT_MIN		10
#define TAB_SIZE		8
#define FILENAME_MAXLEN		255
#define FONT_NAME_MAXLEN	64
#define FONT_SIZE_MAXLEN	3
#define FONT_MAXLEN		FONT_NAME_MAXLEN + 1 + FONT_SIZE_MAXLEN
#define DELIMITER_MAXLEN	80
#define OPTION_NAME_MAXLEN	31
#define OPTION_VALUE_MAXLEN	MAX(MAX(FILENAME_MAXLEN, FONT_MAXLEN), DELIMITER_MAXLEN)
#define OPTION_MAXLEN		OPTION_NAME_MAXLEN + OPTION_VALUE_MAXLEN + 4	/* -name="value" */
#define NOPTIONSMAX		64	/* max number of options */
#define NURLMAX			256	/* max number of url's */
#define NFEEDLINKANDOFFSETMAX	256	/* max number of "open-able" links per feed */
#define LINKTAG_CHAR		'^'	/* char used internally for links, removed from feed text */
#define TMPSTR_SIZE		511	/* used for most tmp strings */
#define FEED_TITLE_MAXLEN	60
#define TITLE_TAG_CHAR		'>'
/*
 * make sure this stuff is robust enough
 */
#define PIXMAP_MAXWIDTH		(32 * 1024 - 1)	/* 32 K - 1 = xpixmaps supported max width */
#define LINE_MAXLEN		128	/* max length of a single line (see news_render.c) */
#define WORD_MAXLEN		32	/* max length of a word that won't get cut */
#if (LINE_MAXLEN <= WORD_MAXLEN)
# error LINE_MAXLEN must be > WORD_MAXLEN
#endif
#define FONT_MAXSIZE		200
/*
 * files and dirs stuff
 *
 * linux version:	/opt/extras.ubuntu.com/news/bin/
 *			/opt/extras.ubuntu.com/news/share/
 *			/opt/extras.ubuntu.com/news/share/pixmaps/
 *			/home/<user_name>/.news/
 *
 * win32 version:	C:\Program files\News
 *			C:\...\Application Data\News
 */
#ifndef G_OS_WIN32
# define NEWSPATH		".news"
#else
# define NEWSPATH		"News"
#endif
#ifndef G_OS_WIN32
# define CONFIGFILE		NEWSPATH"/news_conf"
# define URLLIST		NEWSPATH"/news_url_list"
# define XMLDUMP		NEWSPATH"/news_xml_dump"
# define TMP_XMLDUMP1		NEWSPATH"/tmp_xml_dump1"
# define TMP_XMLDUMP2		NEWSPATH"/tmp_xml_dump2"
# define STD_OUT		stdout
# define STD_ERR		stderr
#else
# define CONFIGFILE		NEWSPATH"\\news_conf"
# define URLLIST		NEWSPATH"\\news_url_list"
# define XMLDUMP		NEWSPATH"\\news_xml_dump"
# define TMP_XMLDUMP1		NEWSPATH"\\tmp_xml_dump1"
# define TMP_XMLDUMP2		NEWSPATH"\\tmp_xml_dump2"
# define STD_OUT		stdout_fp
# define STD_ERR		stderr_fp
#endif
#ifndef G_OS_WIN32
# define INSTALL_DIR		"/opt/extras.ubuntu.com/news"
# define IMAGESPATH		INSTALL_DIR"/share/pixmaps"
#else
# define IMAGESPATH		"News"
#endif
#ifndef G_OS_WIN32
# define NEWS_ICON		IMAGESPATH"/news_icon.png"
# define NEWS_LOGO		IMAGESPATH"/news_logo.png"
# define RSS_ICON		IMAGESPATH"/news_rss_icon.png"
#else
# define NEWS_ICON		IMAGESPATH"\\news_icon.png"
# define NEWS_LOGO		IMAGESPATH"\\news_logo.png"
# define RSS_ICON		IMAGESPATH"\\news_rss_icon.png"
#endif

/*
 * default config values
 */
#define DELAY			8
#define SHIFTSIZE		1
#ifndef G_OS_WIN32
#define FONTNAME		"Sans Bold"
#else
#define FONTNAME		"Verdana Bold"
#endif
#define FONTSIZE		"14"
#define FGCOLOR			"#ffffffff"	/* white */
#define BGCOLOR			"#365162ff"	/* some blue */
#define WIN_X			0
#define WIN_Y			0
#define WIN_W			1024
#define WIN_H			0
#define WINDEC			'n'
#define ALWAYSONTOP		'y'
#define WINTRANSPARENCY		0.8
#define SHADOW			'y'
#define SHADOWOFFSET_X		4
#define SHADOWOFFSET_Y		2
#define SHADOWFX		2
#define LINEDELIMITER		" "
#define CUTLINEDELIMITER	"..."LINEDELIMITER
#define RSSREFRESH		15
#define RSSTITLE		'y'
#define RSSTITLEDELIMITER	"        "
#define RSSDESCRIPTION		'y'
#define RSSDESCRIPTIONDELIMITER	"        -        "
#define STRIPHTMLTAGS		'y'
#define HOMEFEED		"http://newsrss.bbc.co.uk/rss/newsonline_world_edition/front_page/rss.xml"
#ifndef G_OS_WIN32
#define OPENLINKCMD		"firefox"	/* firefox on linux */
#else
#define OPENLINKCMD		""
#endif
#define OPENLINKARGS		""
#define CLOCK			'n'
#ifndef G_OS_WIN32
#define CFONTNAME		"Sans Bold"
#else
#define CFONTNAME		"Verdana Bold"
#endif
#define CFONTSIZE		"12"
#define CFGCOLOR		"#6060ffff"	/* blue */
#define CBGCOLOR		"#000000ff"	/* black */
#define SPECIALCHARS		'n'
#define NEWPG			'`'
#define TABCH			'~'

/*
 * a few usefull graphical (and other) environment values
 */
typedef struct {
	gint		screen_w, screen_h;		/* screen width & height */
	gint		depth;				/* screen depth */
	GtkWidget 	*win;				/* top level window */
	GtkWidget	*drw_a, *drwa_clock;		/* drawing areas (main / clock) */
	gint		drwa_width, drwa_height;	/* main drawing area dimensions */
	GdkDrawable 	*pixmap;			/* pixmap onto which text is rendered */
	gint 		pixm_width, pixm_height;	/* pixmap dimensions */
	GdkGC		*gc;				/* gdk graphical context for pixmap */
	GdkColormap	*colormap;			/* and colormap */

	char		active_link[FILENAME_MAXLEN];	/* then ***not graphical*** params... */
							/* flags below are checked by callback */
	int		suspend_rq;			/* request for doing nothing */
	int		compute_rq;			/* request for (re)computing everything (pixmap and win) */
	int		reload_rq;			/* request for (re)loading resource */
} GrEnv;

typedef struct {
	char		url[FILENAME_MAXLEN + 1];
	gint		offset_in_pixmap;
} FeedLinkAndOffset;

typedef struct {
	int		resrc_type;			/* text file or url */
	char		id[FILENAME_MAXLEN + 1];	/* local = file name or remote = feed url */
	char		feed_title[FEED_TITLE_MAXLEN + 1];	/* obviously only in case it's xml */
	char		xmldump[FILENAME_MAXLEN + 1];	/* obviously only in case it's xml */
	FILE		*fp;				/* file to be read (from xmldump) after some processing, */
							/* not just fopen(file_name) */
	FeedLinkAndOffset link_and_offset[NFEEDLINKANDOFFSETMAX];
} Resource;

typedef enum {
	RESRC_FILE,
	RESRC_URL
} resrc_type;

typedef enum {
	TIME_RESET,
	TIME_CHECK
} check_time_mode;

typedef struct {
	int		delay;
	int		shift_size;
	char		font_n_s[FONT_MAXLEN + 1];
	GdkColor	fg_color;
	guint16		fg_color_alpha;
	GdkColor	bg_color;
	guint16		bg_color_alpha;
	gint		win_x;
	gint		win_y;
	gint		win_w;
	gint		win_h;
	char		windec;
	char		always_on_top;
	gdouble		win_transparency;
	char		shadow;
	int		shadow_offset_x;
	int		shadow_offset_y;
	int		shadow_fx;
	char		line_delimiter[DELIMITER_MAXLEN + 1];
	char		cutline_delimiter[DELIMITER_MAXLEN + 1];
	int		rss_refresh;
	int		rss_title;
	char		rss_title_delimiter[DELIMITER_MAXLEN + 1];
	int		rss_description;
	char		rss_description_delimiter[DELIMITER_MAXLEN + 1];
	char		strip_html_tags;
	char		homefeed[FILENAME_MAXLEN + 1];
	char		open_link_cmd[FILENAME_MAXLEN + 1];
	char		open_link_args[FILENAME_MAXLEN + 1];
	char		clock;
	char		clock_font_n_s[FONT_MAXLEN + 1];
	GdkColor	clock_fg_color;
	guint16		clock_fg_color_alpha;
	GdkColor	clock_bg_color;
	guint16		clock_bg_color_alpha;
	char		special_chars;
	char		new_page_char;
	char		tab_char;
} Params;

typedef enum {
	RESOURCE_OK = LIB_LASTERRORCODE + 1,
	RESOURCE_UNSPECIFIED, RESOURCE_NOT_FOUND, RESOURCE_INVALID,
	RESOURCE_FORMAT_ERROR, RESOURCE_NONE,

	OPTION_OK, OPTION_TOO_MANY, OPTION_ERROR, OPTION_INVALID,
	OPTION_UNKNOWN, OPTION_VALUE_INVALID,

	RSS_OK, RSS_EMPTY, RSS_UNPARSABLE, RSS_CANTCREATEXMLDUMP,
	RSS_CANTREADXMLDUMP, RSS_NO_ITEM_ELEMENT,

	RENDER_OK, RENDER_NO_RESOURCE, RENDER_NULL_ENDINGSTR,
	RENDER_GETLINE_ERROR, RENDER_LINE_TOO_LONG, RENDER_PIXMAP_TOO_WIDE,
	RENDER_OTHER_ERROR,

	LOAD_URL_LIST_OK, LOAD_URL_LIST_ERROR, SAVE_URL_LIST_OK,
	SAVE_URL_LIST_ERROR,

	WIN32V_OK, WIN32V_ERROR,

	WIN32REGKEY_FOUND, WIN32REGKEY_NOT_FOUND, WIN32REGKEY_SAVED,
	WIN32REGKEY_SAVE_ERROR, WIN32REGKEY_CREATE_ERROR,
	WIN32REGKEY_OTHER_ERROR,

	KEY_OK, TIMEOUT_OK, KEY_WRONG, TIMEOUT_WRONG
} news_error_code;

/*
 * news_main.c
 */
void		check_main_win_always_on_top();
void		check_time_load_resource();
GrEnv		*get_gr_env();
Resource	*get_resource();
Params		*get_params();
int		get_instance_id();
gint		update_everything();
/*
 * news_resource.c
 */
int		load_resource(Resource *, int *);
int		format_resource(Resource *);
/*
 * news_render.c
 */
GdkDrawable	*render_stream_to_pixmap(FILE *, FeedLinkAndOffset *, const Params *, int *);
GdkDrawable	*render_stream_to_hpixmap(FILE *, FeedLinkAndOffset *, const Params *, int *);
char		*stream_to_htext(FILE *, char *, PangoLayout *, const Params *, int *);
char 		*void_stream_to_htext(char *, PangoLayout *);
int		get_fsize_from_layout_height(gint, const char *);
gint		get_layout_height_from_fnamesize(const char *);
gint		get_layout_width(PangoLayout *, const char *);
gint		get_layout_height(PangoLayout *, const char *);
/*
 * news_params.c
 */
void		set_default_options(Params *);
int		get_config_file_options(Params *);
int		parse_options_array(Params *, int, const char *[]);
int		get_name_value_from_option(Params *, const char*);
void		save_to_config_file(const Params *);
int		get_gdk_color_from_hexstr(const char *, GdkColor *, guint16 *);
const char	*get_hexstr_from_gdk_color(const GdkColor *, const guint16 *);
void		split_font(const char *, char *, char *);
void		compact_font(char *, const char *, const char *);
const char	*config_file_fullname();
const char	*usr_home_dir();
/*
 * news_clock.c
 */
void		display_time(const Params *);
gint		get_clock_width(const Params *);
/*
 * news_rss.c
 */
int		get_rss_feed(Resource *, const Params *, int *);
int		parse_xml_file(FILE *, const char *, FeedLinkAndOffset *link_and_offset, \
			const Params *);
void		get_selected_elements1(xmlNode *, xmlDoc *);
void		get_selected_elements2(xmlNode *, xmlDoc *, \
			FILE *, FeedLinkAndOffset *link_and_offset, const Params *);
int		get_rss_feed_info(const char*, char *, char *, char *);
void		get_xml_first_element(xmlNode *, xmlDoc *, char *, char *, int);
/*
 * news_list.c
 */
int		load_url_list(char **);
void		sort_url_list(char **);
int		save_url_list(char **);
void		remove_url_from_list(char **, int);
int		str_has_title(char *);
char		*str_without_title(char *);
/*
 * news_rsswin.c
 */
void		get_new_url(Resource *, const Params *, int *);

/*
 * news_prefwin.c
 */
int		modify_params(Params *);
/*
 * news_otherwins.c
 */
void		open_new_txt_file(Resource *resrc, int *);
void		help_win();
void		about_win();
void		info_win(const char *);
void		info_win_wait(const char *, int);
int		question_win(const char *);
int		question_win_always_on_top(const char *);
/*
 * news_misc.c
 */
void		online_help();
void		warning(const char *, const char*, const char*, const char *);
void		dump_font_list();
#ifdef G_OS_WIN32
int 		get_default_browser_from_win32registry(char *);
const char	*get_appdata_dir();
const char	*get_progfiles_dir();
gint		get_win32_taskbar_height();
#endif
void		pango_layout_get_pixel_size2(PangoLayout *, int *, int *);
